import java.io.*;

public class xendecryptFaster
{

	public enum platform{XBOX, PC}
	public static platform myplatform;
	public static void main(String[] args)
	{
	  try
	  {
		  if (args.length != 2)
		  {
			  printUsage();
			  return;
		  }
		  if (!(args[1].endsWith(".fsb.xen")))
		  {
			  System.out.println("File must be in fsb.xen format!");
			  printUsage();
			  return;
		  }
		  if(args[0].equals("-p"))
		  {
			  myplatform = platform.PC;
		  }
		  else if(args[0].equals("-x"))
		  {
			  myplatform = platform.XBOX;
		  }
		  else
		  {
			  System.out.println("Invalid Options\n");
			  printUsage();
			  return;
		  }
		  
		  
		  //Main Code
		  File encryptedName = new File(args[1]);
	      if (encryptedName.length() > (Math.pow(2, 31) -1))
		  {
		    	System.out.println("FSB4 file too large!");
		    	return;
		  }
		  InputStream encryptedStream = new FileInputStream(encryptedName);
		  //NO BYTE BUFFERS!! ONLY INPUTSTREAMS!!
		  System.out.println("Reading File... "+encryptedName.length()+" bytes");
		  byte[] encryptedFile = new byte[(int)encryptedName.length()];
		  encryptedStream.read(encryptedFile);
		  encryptedStream = new FileInputStream(encryptedName);
		  byte[] key = getKey(encryptedFile);
		  if (key == null) 
			  {
			  	keyError();
			  	return;
			  }
		  System.out.println("Decrypting and Writing File...");
		  OutputStream decryptedStream = new FileOutputStream(args[1].substring(0, args[1].length()-4)); 	//".xen" is 4 chars. Since we already checked that
		                         																	  	//it must end in ".xen" we can safely truncate the last 4 chars
		  decrypt(encryptedStream, key, decryptedStream);
		  OutputStream keyStream = new FileOutputStream((args[1].substring(0, args[1].length()-8)).concat("_keyfile.bin")); //".fsb.xen" is 8 chars.
		  //System.out.println("Writing File...");
		  //decryptedFile.write(decryptedBuffer.array());
		  System.out.println("Writing KeyFile.bin");
		  keyStream.write(key);
		  keyStream.close();
		  decryptedStream.close();
		  encryptedStream.close();
	  }
	  catch (IOException e)
	  {
		  System.err.println("Caught IO Exception: cannot read/write data!: " + e.getMessage());
	  }
	}
	private static void printUsage()
	{
		System.out.println("Usage...");
		System.out.println("java xendecrypt [options] <song.fsb.xen>");
		System.out.println("Options: -p PC   FSB.xen");
		System.out.println("         -x XBOX FSB.xen");
		System.out.println("");
		System.out.println("Outputs song.fsb and song_keyfile.bin for future use.");
	}
	private static byte[] getKey(byte[] encryptedFile)
	{
		boolean keyFound = false;
		int keylen = 1;
		/*
	  	byte[] id =   {70, 83, 66, 52}; //"FSB"
	  	byte[] numsamples = {1,  0,  0,  0 }; //Num Samples is always 1
	  	byte[] shdrsize =  {0,  0,  80, 0 }; //Always 80 bytes long.
	  	byte[] datasize = new byte[4]; //Don't know this yet, the number is the file size - 128 (the size of the headers) laid out in little-endian.
	  	byte[] version =  {0,  0,  4,  0 }; //Always 0x00040000 for FSB4.
	  	byte[] mode =       {0,  0,  0,  0 }; //I think this is always 0x00000000 mode 0.
	  	byte[] zero =   {0,  0,  0,  0,  0,  0,  0,  0 }; //Just 8 bytes of zeroes. Nothing special
		 */
		//Look for repeated stuff. 0000's or 1111's.
		//Set datasize byte array
		if (myplatform == platform.PC){
			byte[] datasize = new byte[4];
			byte[] headerPCCheck = {70, 83, 66, 52,
		         1,  0,  0,  0,
		         80, 0,  0,  0,
		         0,  0,  0,  0,
		         0,  0,  4,  0,
		         0,  0,  0,  0,
		         0,  0,  0,  0,
		         0,  0,  0,  0};
			int encryptedDataSize = encryptedFile.length-128;
			int datacarry = encryptedDataSize;
			//lay out data in little-endian format.
			for(int i=3; i>-1; i--)
			{
				datasize[i] = (byte) (datacarry / (Math.pow(16, i*2)));
				datacarry -= ((datacarry / (int) (Math.pow(16, i*2)))* (int) Math.pow(16, i*2));
			}
			System.arraycopy(datasize, 0, headerPCCheck, 12, 4); //Copy new data into headerCheck for key generation.
			while(!keyFound)
			{  
				if (keylen == 32)
				{
					keyError();
					return null;
				}	
				//Start with key of length 1. Check with file header, from fsb.h. FSB file header is 128 bytes, 48 byte header, with 80 byte sample header. 
				//We can read keys up to length of 31, because we only check in the first 32 bytes for the header, the last 16 is the hash, which i have no idea about..!!
				//2A xor 46 = 6C -> (swap) 36
				//2A xor 48 = 62 (01100010) -> (swap) 46 (01000110)
				//62 xor 48 = 2A
				//Swap byte bits of headerCheck, then xor with encrypteddata to get key.
				byte[] key = new byte[keylen];
				for(int i=0; i<keylen; i++)
				{	
					key[i] = (byte) (swapByteBits(headerPCCheck[i])^encryptedFile[i]);
				}
				//Now we can procede and examine if this key is valid.
				int checkMax;
				if (keylen < (32 - keylen))
				{
					checkMax = keylen;
				}
				else 
					checkMax = (32-keylen);
				for(int i=0; i<checkMax; i++)
				{
					if (headerPCCheck[i+keylen] != swapByteBits((byte) (encryptedFile[i+keylen] ^ key[i%key.length])))
					{
						keylen++;
						break;
					}
					if (i == (checkMax-1))
					{
						keyFound = true;
						return key;
					}
				}
				//if it gets through the error checking routine, we have a working key!
			}
		}
		else //platform == XBOX
		{
			byte[] datasize = new byte[4];
			byte[] headerXBCheck1 = {70, 83, 66, 52,
			         1,  0,  0,  0,
			         0,  0,  0,  0, /*dunno this part at all*/
			         0,  0,  0,  0, /*this is filesize-lastpart*/
			         0,  0,  4,  0,
			         0,  0,  0,  0,
			         0,  0,  0,  0,
			         0,  0,  0,  0};
			byte[] headerXBCheck2 = {0x6D, 0x75, 
			         0x6C, 0x74, 0x69,  0x63,
			         0x68, 0x61, 0x6E,  0x6E,
			         0x65, 0x6C, 0x20,  0x73,
			         0x6F, 0x75, 0x6E,  0x64,
			         0,  0,  0,  0,
			         0,  0,  0,  0,
			         0,  0,  0,  0};
			//set encryptedDataSize...
			while(!keyFound)
			{  
				if (keylen == 32)
				{
					keyError();
					return null;
				}
				//Copy missing bytes into headerCheck...
				//Indexes 50-79...
				int header2index = ((50/keylen)*keylen + 8);
				if (header2index < 50)
				{
					header2index += keylen;
				}
				for(int i=8; i<12; i++)
				{
					headerXBCheck1[i] = (byte) swapByteBits((byte) ((swapByteBits(headerXBCheck2[header2index-50])^(encryptedFile[header2index+i-8]))^(encryptedFile[i])));
				}
				//swap((swap(plain) xor cipher) xor cipher) = plain.
		   	  	byte[] dataOffset = new byte[4];
		   	  	System.arraycopy(headerXBCheck1, 8, dataOffset, 0, 4);
		   	  	//DataOffset might have negative values!!!!
		   	  	int dataOffsetCarry=0;
		   	  	for(int i=3; i>-1; i--)
		   	  	{
		   	  		int getByte = dataOffset[i];
		   	  		if (getByte < 0) //if negative convert to positive!
		   	  		{
		   	  			getByte += 256;
		   	  		}
		   	  		dataOffsetCarry += (getByte * Math.pow(16, i*2));
		   	  	}
		   	  	int encryptedDataSize = encryptedFile.length-dataOffsetCarry-48;
		   	  	int datacarry = encryptedDataSize;
		   	  	//lay out data in little-endian format.
		   	  	for(int i=3; i>-1; i--)
		   	  	{
		   	  		int cursub = (datacarry / (int) (Math.pow(16, i*2)));
		   	  		datasize[i] = (byte) cursub;
		   	  		datacarry -= (cursub* (int) Math.pow(16, i*2));
		   	  	}
		   	  	System.arraycopy(datasize, 0, headerXBCheck1, 12, 4);
		   	  	byte[] key = new byte[keylen];
		   	  	for(int i=0; i<keylen; i++)
		   	  	{
		   	  		key[i] = (byte) (swapByteBits(headerXBCheck1[i])^encryptedFile[i]);
		   	  	}
	
		   	  	//Now we can procede and examine if this key is valid.
		   	  	int checkMax;
		   	  	if (keylen < (32 - keylen))
		   	  	{
		   	  		checkMax = keylen;
		   	  	}
		   	  	else checkMax = (32-keylen);
		   	  	for(int i=0; i<checkMax; i++)
		   	  	{
		   	  		if (headerXBCheck1[i+keylen] != swapByteBits((byte) (encryptedFile[i+keylen] ^ key[i%key.length])))
		   	  		{
		   	  			keylen++;
		   	  			break;
		   	  		}
		   	  		if (i == (checkMax-1))
		   	  		{
		   	  			keyFound = true;
		   	  			return key;
		   	  		}
		   	  	}
		   	  	//if it gets through the error checking routine, we have a working key!
			}
		}
		return null;
	}
	private static void keyError() {
		System.out.println("ERROR01: Key not found, or Input File not correct format!");
	}
	private static byte swapByteBits(byte cInput)
	{
   
		byte nResult = 0;
		for(int i=0; i<8; i++)
		{
			nResult = (byte) (nResult << 1);
			nResult |= (cInput&1);
			cInput = (byte) (cInput >> 1);
		} 
		return nResult;
	}
 
	private static void decrypt(InputStream encryptedStream, byte[] key, OutputStream decryptedStream) throws IOException
	{
        byte [] swapTab = new byte [256];
		boolean eofReached = false;

        for (int i=0;i<256;i++)
        {
            swapTab[i] = swapByteBits((byte)i);
        }
        byte [] inbuf = new byte [2048];
        byte [] outbuf = new byte [2048];
		int i = 0;
		while(!eofReached)
		{
			int bsize = encryptedStream.read(inbuf);
			if (bsize == -1)
			{
				eofReached = true;
				break;
			}
            for (int j = 0; j < bsize; j++)
            {
                outbuf[j] = swapTab[255 & (inbuf[j]^key[i % key.length])];
			    i++;
            }
			decryptedStream.write(outbuf,0,bsize);
		}
	}
}
